// Copyright 2025 The Accumulate Authors
//
// Use of this source code is governed by an MIT-style
// license that can be found in the LICENSE file or at
// https://opensource.org/licenses/MIT.

package testing

import (
	"crypto/ecdsa"
	"crypto/ed25519"
	"crypto/sha256"
	"encoding/binary"
	"fmt"
	"math/big"
	badrand "math/rand"
	"strings"

	btc "github.com/btcsuite/btcd/btcec"
	tmcrypto "github.com/cometbft/cometbft/crypto"
	tmed25519 "github.com/cometbft/cometbft/crypto/ed25519"
	"gitlab.com/accumulatenetwork/accumulate/internal/database"
	"gitlab.com/accumulatenetwork/accumulate/internal/database/indexing"
	"gitlab.com/accumulatenetwork/accumulate/internal/database/smt/storage"
	"gitlab.com/accumulatenetwork/accumulate/pkg/build"
	"gitlab.com/accumulatenetwork/accumulate/pkg/types/messaging"
	"gitlab.com/accumulatenetwork/accumulate/pkg/types/record"
	"gitlab.com/accumulatenetwork/accumulate/pkg/url"
	"gitlab.com/accumulatenetwork/accumulate/protocol"
)

type DB = *database.Batch

func GenerateKey(seed ...interface{}) ed25519.PrivateKey {
	h := storage.MakeKey(seed...)
	return ed25519.NewKeyFromSeed(h[:])
}

func GenerateTmKey(seed ...interface{}) tmed25519.PrivKey {
	return tmed25519.PrivKey(GenerateKey(seed...))
}

// generates privatekey and compressed public key
func NewSECP256K1(seed ...any) *btc.PrivateKey {
	hash := record.NewKey(seed...).Hash()
	src := badrand.NewSource(int64(binary.BigEndian.Uint64(hash[:])))
	priv, err := ecdsa.GenerateKey(btc.S256(), badrand.New(src))
	if err != nil {
		panic(err)
	}
	return (*btc.PrivateKey)(priv)
}

func ParseUrl(s string) (*url.URL, error) {
	u, err := url.Parse(s)
	if err != nil {
		return nil, err
	}
	if protocol.AcmeUrl().Equal(u) {
		return u, nil
	}
	if key, _ := protocol.ParseLiteAddress(u); key != nil {
		return u, nil
	}
	if _, err := protocol.ParseLiteDataAddress(u); err == nil {
		return u, nil
	}
	// Fixup URL
	if !strings.HasSuffix(u.Authority, protocol.TLD) {
		u = u.WithAuthority(u.Authority + protocol.TLD)
	}
	return u, nil
}

func BuildTestTokenTxGenTx(sponsor ed25519.PrivateKey, destAddr string, amount uint64) (*messaging.Envelope, error) {
	//use the public key of the bvc to make a sponsor address (this doesn't really matter right now, but need something so Identity of the BVC is good)
	from := AcmeLiteAddressStdPriv(sponsor)

	u, err := ParseUrl(destAddr)
	if err != nil {
		return nil, fmt.Errorf("invalid URL: %v", err)
	}

	send := protocol.SendTokens{}
	send.AddRecipient(u, big.NewInt(int64(amount)))

	return build.Transaction().
		For(from).
		Body(&send).
		SignWith(from.RootIdentity()).Version(1).Timestamp(1).PrivateKey(sponsor).Type(protocol.SignatureTypeLegacyED25519).
		Done()
}

func CreateLiteTokenAccount(db DB, key tmed25519.PrivKey, tokens float64) error {
	url := AcmeLiteAddressTmPriv(key)
	err := CreateTokenAccount(db, url.String(), protocol.AcmeUrl().String(), tokens, true)
	if err != nil {
		return err
	}
	liteIdUrl := url.RootIdentity()
	return CreateLiteIdentity(db, liteIdUrl.String(), 0)
}

func AddCredits(db DB, account *url.URL, credits float64) error {
	state, err := db.Account(account).Main().Get()
	if err != nil {
		return err
	}
	switch state.Type() {
	case protocol.AccountTypeLiteIdentity, protocol.AccountTypeKeyPage:
	default:
		return fmt.Errorf("%s does not refer to a lite identity", account.String())
	}

	state.(protocol.AccountWithCredits).CreditCredits(uint64(credits * protocol.CreditPrecision))
	return db.Account(account).Main().Put(state)
}

func CreateLiteTokenAccountWithCredits(db DB, key tmed25519.PrivKey, tokens, credits float64) error {
	url := AcmeLiteAddressTmPriv(key)
	err := CreateTokenAccount(db, url.String(), protocol.AcmeUrl().String(), tokens, true)
	if err != nil {
		return err
	}
	liteIdUrl := url.RootIdentity()
	return CreateLiteIdentity(db, liteIdUrl.String(), credits)
}

func WriteStates(db DB, chains ...protocol.Account) error {
	txid := sha256.Sum256([]byte("fake txid"))
	urls := make([]*url.URL, len(chains))
	for i, c := range chains {
		urls[i] = c.GetUrl()

		r := db.Account(c.GetUrl())
		err := r.Main().Put(c)
		if err != nil {
			return err
		}

		chain, err := r.MainChain().Get()
		if err != nil {
			return err
		}

		err = chain.AddEntry(txid[:], true)
		if err != nil {
			return err
		}
	}

	dir := indexing.Directory(db, urls[0].Identity())
	var uarr []*url.URL
	for _, u := range urls {
		if u.IsRootIdentity() {
			continue
		}
		uarr = append(uarr, u)
	}
	err := dir.Add(uarr...)
	if err != nil {
		return err
	}
	return nil
}

func CreateADI(db DB, key tmed25519.PrivKey, urlStr string) error {
	keyHash := sha256.Sum256(key.PubKey().Bytes()) // TODO This is not what create_identity / create_key_page do, nonce will be > 0 also

	identityUrl, err := ParseUrl(urlStr)
	if err != nil {
		return err
	}

	bookUrl := identityUrl.JoinPath("book0")

	ss := new(protocol.KeySpec)
	ss.PublicKeyHash = keyHash[:]

	page := new(protocol.KeyPage)
	page.Url = protocol.FormatKeyPageUrl(bookUrl, 0)
	page.AddKeySpec(ss)
	page.AcceptThreshold = 1
	page.Version = 1

	book := new(protocol.KeyBook)
	book.Url = bookUrl
	book.AddAuthority(bookUrl)
	book.PageCount = 1

	adi := new(protocol.ADI)
	adi.Url = identityUrl
	adi.AddAuthority(bookUrl)

	return WriteStates(db, adi, book, page)
}

func CreateSubADI(db DB, originUrlStr string, urlStr string) error {
	originUrl, err := ParseUrl(originUrlStr)
	if err != nil {
		return err
	}
	identityUrl, err := ParseUrl(urlStr)
	if err != nil {
		return err
	}

	adi := new(protocol.ADI)
	adi.Url = identityUrl
	adi.AddAuthority(originUrl.JoinPath("book0"))

	return WriteStates(db, adi)
}

func CreateAdiWithCredits(db DB, key tmed25519.PrivKey, urlStr string, credits float64) error {
	err := CreateADI(db, key, urlStr)
	if err != nil {
		return err
	}

	u, err := ParseUrl(urlStr)
	if err != nil {
		return err
	}

	return AddCredits(db, u.JoinPath("book0/1"), credits)
}

func CreateLiteIdentity(db DB, accUrl string, credits float64) error {
	u, err := ParseUrl(accUrl)
	if err != nil {
		return err
	}

	var chain protocol.Account
	account := new(protocol.LiteIdentity)
	account.Url = u
	account.CreditBalance = uint64(credits * protocol.CreditPrecision)
	chain = account
	return db.Account(u).Main().Put(chain)
}

func CreateTokenAccount(db DB, accUrl, tokenUrl string, tokens float64, lite bool) error {
	u, err := ParseUrl(accUrl)
	if err != nil {
		return err
	}

	tu, err := ParseUrl(tokenUrl)
	if err != nil {
		return err
	}

	var chain protocol.Account
	if lite {
		account := new(protocol.LiteTokenAccount)
		account.Url = u
		account.TokenUrl = tu
		account.Balance.SetInt64(int64(tokens * protocol.AcmePrecision))
		chain = account
	} else {
		account := new(protocol.TokenAccount)
		account.Url = u
		account.TokenUrl = tu
		account.AddAuthority(u.Identity().JoinPath("book0"))
		account.Balance.SetInt64(int64(tokens * protocol.AcmePrecision))
		chain = account
	}

	return db.Account(u).Main().Put(chain)
}

func CreateTokenIssuer(db DB, urlStr, symbol string, precision uint64, supplyLimit *big.Int) error {
	u, err := ParseUrl(urlStr)
	if err != nil {
		return err
	}

	issuer := new(protocol.TokenIssuer)
	issuer.Url = u
	issuer.AddAuthority(u.Identity().JoinPath("book0"))
	issuer.Symbol = symbol
	issuer.Precision = precision
	issuer.SupplyLimit = supplyLimit

	return db.Account(u).Main().Put(issuer)
}

func CreateKeyPage(db DB, bookUrlStr string, keys ...tmed25519.PubKey) error {
	bookUrl, err := ParseUrl(bookUrlStr)
	if err != nil {
		return err
	}

	account := db.Account(bookUrl)
	state, err := account.Main().Get()
	if err != nil {
		return err
	}
	book := state.(*protocol.KeyBook)

	page := new(protocol.KeyPage)
	page.Url = protocol.FormatKeyPageUrl(bookUrl, book.PageCount)
	page.AcceptThreshold = 1
	page.Version = 1
	page.Keys = make([]*protocol.KeySpec, 0, len(keys))
	for _, key := range keys {
		hash := sha256.Sum256(key.Bytes())
		page.AddKeySpec(&protocol.KeySpec{
			PublicKeyHash: hash[:],
		})
	}
	book.PageCount++
	return WriteStates(db, page, book)
}

func CreateKeyBook(db DB, urlStr string, publicKey tmed25519.PubKey) error {
	bookUrl, err := ParseUrl(urlStr)
	if err != nil {
		return err
	}

	book := new(protocol.KeyBook)
	book.Url = bookUrl
	book.AddAuthority(bookUrl)
	book.PageCount = 1

	page := new(protocol.KeyPage)
	page.Version = 1
	page.Url = protocol.FormatKeyPageUrl(bookUrl, 0)

	key := new(protocol.KeySpec)
	hash := sha256.Sum256(publicKey)
	key.PublicKeyHash = hash[:]
	page.Keys = []*protocol.KeySpec{key}

	accounts := []protocol.Account{book, page}
	return WriteStates(db, accounts...)
}

func CreateAccount(db DB, account protocol.FullAccount) error {
	auth := account.GetAuth()
	if len(auth.Authorities) > 0 {
		return WriteStates(db, account)
	}

	var identity *protocol.ADI
	err := db.Account(account.GetUrl().Identity()).Main().GetAs(&identity)
	if err != nil {
		return err
	}

	*auth = *identity.GetAuth()
	return WriteStates(db, account)
}

func UpdateKeyPage(db DB, account *url.URL, fn func(*protocol.KeyPage)) error {
	return UpdateAccount(db, account, fn)
}

func UpdateAccount[T protocol.Account](db DB, accountUrl *url.URL, fn func(T)) error {
	var account T
	err := db.Account(accountUrl).Main().GetAs(&account)
	if err != nil {
		return err
	}

	fn(account)
	return db.Account(accountUrl).Main().Put(account)
}

// AcmeLiteAddress creates an ACME lite address for the given key. FOR TESTING
// USE ONLY.
func AcmeLiteAddress(pubKey []byte) *url.URL {
	u, err := protocol.LiteTokenAddress(pubKey, protocol.ACME, protocol.SignatureTypeED25519)
	if err != nil {
		// LiteTokenAddress should only return an error if the token URL is invalid,
		// so this should never return an error. But ignoring errors is a great
		// way to get bugs.
		panic(err)
	}
	return u
}

func AcmeLiteAddressTmPriv(key tmcrypto.PrivKey) *url.URL {
	return AcmeLiteAddress(key.PubKey().Bytes())
}

func AcmeLiteAddressStdPriv(key ed25519.PrivateKey) *url.URL {
	return AcmeLiteAddress(key[32:])
}
