// Copyright 2025 The Accumulate Authors
//
// Use of this source code is governed by an MIT-style
// license that can be found in the LICENSE file or at
// https://opensource.org/licenses/MIT.

package simulator

import (
	"encoding/json"
	"testing"

	"github.com/stretchr/testify/require"
	"gitlab.com/accumulatenetwork/accumulate/pkg/types/encoding"
	"gitlab.com/accumulatenetwork/accumulate/pkg/types/messaging"
	"golang.org/x/exp/rand"
)

func TestDeterministicMessageOrder(t *testing.T) {
	list := &encoding.JsonUnmarshalListWith[messaging.Message]{Func: messaging.UnmarshalMessageJSON}
	require.NoError(t, json.Unmarshal([]byte(messageData), list))

	sorted := list.Value
	orderMessagesDeterministically(sorted)

	for i := 0; i < 100; i++ {
		// Copy the array and shuffle it
		messages := make([]messaging.Message, len(sorted))
		copy(messages, sorted)
		rand.Shuffle(len(sorted), func(i, j int) {
			messages[i], messages[j] = messages[j], messages[i]
		})

		// Sort
		orderMessagesDeterministically(messages)

		// Verify the resulting order matches
		for i := range sorted {
			if !messaging.EqualMessage(sorted[i], messages[i]) {
				b, _ := json.Marshal(sorted[i])
				c, _ := json.Marshal(messages[i])
				require.Truef(t, false, "Mismatch at %d:\n%s\n%s", i, b, c)
			}
		}
	}
}

// messageData is a big pile of messages extracted from a test. It's content is pretty arbitrary.
const messageData = `[
	{"type":"transaction","transaction":{"header":{"principal":"acc://alice/tokens","initiator":"1e92229dfb615cf2fc08a9c2a3c22f98e8b8741553e69e59886c26e7adb3d8a2"},"body":{"type":"sendTokens","hash":"0000000000000000000000000000000000000000000000000000000000000000","to":[{"url":"acc://bob/tokens","amount":"123"}]}}},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"48892b8f08e6cd764f70f1268974a70e271c9fe751afd6052e63ec05f7340e21","signature":"eaf2392eef81efcf770dc7fcd571b941f39cdcf46658b7f5d9ac59951f2adc10139f1653cf69feb1a27dc3066777f18db19ab9b2be51d46c756f51983966660d","signer":"acc://alice/book/1","signerVersion":1,"timestamp":1,"transactionHash":"0a46b8493ba8f208d37d02bd0c11fe1ac5787c7924b4725f26c4c689e5a6b904"},"transactionHash":"0a46b8493ba8f208d37d02bd0c11fe1ac5787c7924b4725f26c4c689e5a6b904"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"c1ffe952729abc7af63f381daf62de7740bfe38ec80934b4fced75bba6fbabfd"},"body":{"type":"blockValidatorAnchor","source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2","acmeBurnt":"0"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://bvn-BVN2.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":1,"transactionHash":"d693bdca4ee6ce65bb94b150fc9940a31d4e6eff810d25b154606d327c361be0"},"transactionHash":"d693bdca4ee6ce65bb94b150fc9940a31d4e6eff810d25b154606d327c361be0"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"9de8f58234a823c9b9618d852accb42559c0b16ae225f9bdbe9993731a4b6d5c","signature":"50a3070a6c665b06b8e451c4df5063ded736911f55ffbf2cc990454455424caf7bedd518fcd19bd91ab3889bd72733b538230b9c7d98ba8259fd97b718a8f505","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"d693bdca4ee6ce65bb94b150fc9940a31d4e6eff810d25b154606d327c361be0"},"transactionHash":"d693bdca4ee6ce65bb94b150fc9940a31d4e6eff810d25b154606d327c361be0"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"c1ffe952729abc7af63f381daf62de7740bfe38ec80934b4fced75bba6fbabfd"},"body":{"type":"blockValidatorAnchor","source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2","acmeBurnt":"0"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://bvn-BVN2.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":1,"transactionHash":"d693bdca4ee6ce65bb94b150fc9940a31d4e6eff810d25b154606d327c361be0"},"transactionHash":"d693bdca4ee6ce65bb94b150fc9940a31d4e6eff810d25b154606d327c361be0"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"78641be5bc680f8247265bf2da3cd31afd111c3f59379d80f095986d5fedd07a","signature":"ce8eef8aece1235843a9e670a1e09800f9a7fd3611334958f8a6954fd37fa1e880bf6658b4500aa909a993551099d3a015397b981e2711dc235db485e737bb03","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"d693bdca4ee6ce65bb94b150fc9940a31d4e6eff810d25b154606d327c361be0"},"transactionHash":"d693bdca4ee6ce65bb94b150fc9940a31d4e6eff810d25b154606d327c361be0"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"c1ffe952729abc7af63f381daf62de7740bfe38ec80934b4fced75bba6fbabfd"},"body":{"type":"blockValidatorAnchor","source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2","acmeBurnt":"0"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://bvn-BVN2.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":1,"transactionHash":"d693bdca4ee6ce65bb94b150fc9940a31d4e6eff810d25b154606d327c361be0"},"transactionHash":"d693bdca4ee6ce65bb94b150fc9940a31d4e6eff810d25b154606d327c361be0"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"05f3fcc7e95c17eaab6692d673176249f114a16bc28d2996eb5c08f6c6880320","signature":"adc87e3c0216595cc3b81de919e88f24e2ee599285701180cd7cc56477364303d71f9fc2f47e23366bdfd206b04963a268b3471eb672c1f92b7ccd960a765605","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"d693bdca4ee6ce65bb94b150fc9940a31d4e6eff810d25b154606d327c361be0"},"transactionHash":"d693bdca4ee6ce65bb94b150fc9940a31d4e6eff810d25b154606d327c361be0"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"1e17fd8766c93486299998719ff9bf0f76474e1acf729b4a32cb19f4d960c84a"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":1,"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"b47dc38466953fa79e81abfc7bd19d564f870ec07f26aa1fb06751ea4bf7c4b1","signature":"9e4af685f1c16553b14c2d432d9fbf532a5a1be38f1d06a4194a6dc10d966092829b5a5d64c2d610fd8f60e72a80d116a1197eeec9a5c9c32baf384d1ab96704","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN0.acme/anchors","initiator":"d8b499a53e6a0080678dd2f18b0f9fecd587860c8c797df0b8730ff4bc0d8574"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN0.acme","sequenceNumber":1,"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"b47dc38466953fa79e81abfc7bd19d564f870ec07f26aa1fb06751ea4bf7c4b1","signature":"67ce56fa5c1932f4f735f1f53e15641f7767cf9387e7d5b441f05ffbe3bae261dbcdda8fa1c0c8797fda513aca5dc10419e8deacb87fc7ef98a63c5d3b656b0b","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN1.acme/anchors","initiator":"612f2c9a306dc213e6f81781af1f1bb7b16b3d1d50f23ea1230663a22cbae03c"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN1.acme","sequenceNumber":1,"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"b47dc38466953fa79e81abfc7bd19d564f870ec07f26aa1fb06751ea4bf7c4b1","signature":"e1d54389256abd75c4095cc2b5937cd675aadd704a1855fce6dc5c29cf1f8af02918af93bd32a8687063e84e2ceb444689d44b9c852df9a88aa1d8bf96ae4b00","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN2.acme/anchors","initiator":"853eb946ff51dc1dfb92b6dab1215e84c551da1bd465e3e82d2b38a51989e81c"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN2.acme","sequenceNumber":1,"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"b47dc38466953fa79e81abfc7bd19d564f870ec07f26aa1fb06751ea4bf7c4b1","signature":"d249cdf3d9f0ff0e7e8e496a959f583cd23c4b6e47b9bee4e8656617b1bbc722819581ccceb1098c12b46e7daa289574864b6a8ddf397ec2e6e5c2e017301d01","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN1.acme/anchors","initiator":"612f2c9a306dc213e6f81781af1f1bb7b16b3d1d50f23ea1230663a22cbae03c"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN1.acme","sequenceNumber":1,"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"f0ba4305a861d7a3c1869f2632886e07134a03115f182f860b1113afd0121438","signature":"62a6ea5c5db7ab524501c0c1534655b4ee5e1653386eb44ea00e82aaee5ed868de1c2d2f1d5ede4729a80a9fa85a057c796f919d2e5f1d9949a439b22c904304","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN2.acme/anchors","initiator":"853eb946ff51dc1dfb92b6dab1215e84c551da1bd465e3e82d2b38a51989e81c"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN2.acme","sequenceNumber":1,"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"f0ba4305a861d7a3c1869f2632886e07134a03115f182f860b1113afd0121438","signature":"3d9a836f1b1e600ee82616a09eed393f4c8a46e1721ac50c6b573032d63884c4837adb690003df31a067426ab121783a2ab8c983347a40124f7153c799e78f09","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"1e17fd8766c93486299998719ff9bf0f76474e1acf729b4a32cb19f4d960c84a"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":1,"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"f0ba4305a861d7a3c1869f2632886e07134a03115f182f860b1113afd0121438","signature":"b53a88f079fc6eb64f9db99225f474f8f39aae3521904f57ba0b9ff2ec7e05b1d12182b71fab314b3eb1b180f55456878d00113a63e7be8fe41fdfcf9d78ec06","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN0.acme/anchors","initiator":"d8b499a53e6a0080678dd2f18b0f9fecd587860c8c797df0b8730ff4bc0d8574"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN0.acme","sequenceNumber":1,"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"f0ba4305a861d7a3c1869f2632886e07134a03115f182f860b1113afd0121438","signature":"c4e21c4880e9a92554a1c8be6b402fc3cc02765d410dd6d118fa30ed362801095d6a4e14c2003533194a38f0c755958a44dd7a79a4b5cd5f8d3db586a373fd0f","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"1e17fd8766c93486299998719ff9bf0f76474e1acf729b4a32cb19f4d960c84a"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":1,"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"60e7be68143da64401f57be2fe88640d06c0d314fea381ec7dc5a21e7516d758","signature":"ecf550703bcdcafcfe05457c15c5a41052b63a4145b51c80933ecbdd6d2cf1ffeac519e7aeebc7ffe8f50448098af73664f274c5d4e27ea7ca73f02d7c9aac0b","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN0.acme/anchors","initiator":"d8b499a53e6a0080678dd2f18b0f9fecd587860c8c797df0b8730ff4bc0d8574"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN0.acme","sequenceNumber":1,"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"60e7be68143da64401f57be2fe88640d06c0d314fea381ec7dc5a21e7516d758","signature":"69b68aea62bd6bfd0f9e9f0f41ceda9932a484113274cdb7b966c4cd43baf1f4bbe1ecf134aa510ae692d89df7e3d6fa931afce13aea272218740e6ec9140f03","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN1.acme/anchors","initiator":"612f2c9a306dc213e6f81781af1f1bb7b16b3d1d50f23ea1230663a22cbae03c"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN1.acme","sequenceNumber":1,"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"60e7be68143da64401f57be2fe88640d06c0d314fea381ec7dc5a21e7516d758","signature":"c2b9f9fbbc1d9493190c4ffb855ff2767f8084e1f44831d1d89b7b6933f9d98492a369193fc1fc4f15c83b126d5adae004a053fd7ba5bf2e7c43cf1e7736d00e","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN2.acme/anchors","initiator":"853eb946ff51dc1dfb92b6dab1215e84c551da1bd465e3e82d2b38a51989e81c"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN2.acme","sequenceNumber":1,"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"60e7be68143da64401f57be2fe88640d06c0d314fea381ec7dc5a21e7516d758","signature":"c64428b4eeaffb700e7a5aea7068e9794a3896b87f030d759d4d04a4cbfceb40db398c9d0b9e1482ffe504bad328118fa935d12649a38f5026f730d937c4d80b","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN0.acme/anchors","initiator":"d8b499a53e6a0080678dd2f18b0f9fecd587860c8c797df0b8730ff4bc0d8574"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN0.acme","sequenceNumber":1,"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"59d55ae3c2c644bf8c849d451eb8733b47f219850157decf2b7f3a5cd3c96c6b","signature":"59b785b2deb7f983d59c3f468e6367c7d9b03e964c6c56f82497c70df4acd9c1cd89f0c4bd1777f020218fab670eb17569190afd043e5c6ea4b6350fd3c75305","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN1.acme/anchors","initiator":"612f2c9a306dc213e6f81781af1f1bb7b16b3d1d50f23ea1230663a22cbae03c"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN1.acme","sequenceNumber":1,"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"59d55ae3c2c644bf8c849d451eb8733b47f219850157decf2b7f3a5cd3c96c6b","signature":"a87e2d5be607c5a29828ddeb7e38912a2faad8f49115ece92b9aab37345c3d15d9541e4fd5124cdc39f3d624b8d4b0c8dbb36651139dd0855e9123ebfa982401","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN2.acme/anchors","initiator":"853eb946ff51dc1dfb92b6dab1215e84c551da1bd465e3e82d2b38a51989e81c"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN2.acme","sequenceNumber":1,"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"59d55ae3c2c644bf8c849d451eb8733b47f219850157decf2b7f3a5cd3c96c6b","signature":"d3f64cf9b489b25ab181b24540d83f9979ef636b2a7aaaf888e41153be76b12c51f88d393b2cc3294a2c5c3ed9027f0dafd3374206a7d87f5fa0cde64b7a7f04","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"1e17fd8766c93486299998719ff9bf0f76474e1acf729b4a32cb19f4d960c84a"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":1,"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"59d55ae3c2c644bf8c849d451eb8733b47f219850157decf2b7f3a5cd3c96c6b","signature":"d3520ee9ea80032f5f034cf25ab9d375c51d8fc0fa3a4dbcf3c1e56b62ba32ce212a1012ca8386702a1371f669a4c0f7cd336e49bc2450c384e67283d3ce4002","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN2.acme/anchors","initiator":"853eb946ff51dc1dfb92b6dab1215e84c551da1bd465e3e82d2b38a51989e81c"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN2.acme","sequenceNumber":1,"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"1a7551ffe00adda17a6c1af8f602dbdb4bdd7cbdc96f33cab1e938a2756f0e5d","signature":"cf2a7fa0de35641954f2207c6509b7ea5a0ed64002bd7a8dbf1b76dbec456bcfddb47a3967b85ae6137a3a16ea93189fdff12266b3197ad560820c0cd66e2b0f","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"1e17fd8766c93486299998719ff9bf0f76474e1acf729b4a32cb19f4d960c84a"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":1,"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"1a7551ffe00adda17a6c1af8f602dbdb4bdd7cbdc96f33cab1e938a2756f0e5d","signature":"af4e66741609bb4d337e2214c5907f73081c49ad5e614326a5dc04ad429df20c20f31b09713ed61a57f3fcdaba982e64d7925ebfd1648df892fe27d63a29980f","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN0.acme/anchors","initiator":"d8b499a53e6a0080678dd2f18b0f9fecd587860c8c797df0b8730ff4bc0d8574"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN0.acme","sequenceNumber":1,"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"1a7551ffe00adda17a6c1af8f602dbdb4bdd7cbdc96f33cab1e938a2756f0e5d","signature":"24f5a8744851c9a37a2b72ed785c23d3470058105daaafe4110fbb51ef61c5fc800d72962f585a7f8b51825bc2f9ea37aebbfe2a0f00b9e526b3e084c62ae902","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN1.acme/anchors","initiator":"612f2c9a306dc213e6f81781af1f1bb7b16b3d1d50f23ea1230663a22cbae03c"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN1.acme","sequenceNumber":1,"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"1a7551ffe00adda17a6c1af8f602dbdb4bdd7cbdc96f33cab1e938a2756f0e5d","signature":"2676027b9de5b271a6668b725c0c1cf0ba8d66e3f0fcf6300bde2280a76bf17bfebbd3269df3781f4e2ba06bbd803a469645ab65777b2903c2291ece96c37309","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN0.acme/anchors","initiator":"d8b499a53e6a0080678dd2f18b0f9fecd587860c8c797df0b8730ff4bc0d8574"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN0.acme","sequenceNumber":1,"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"633ef8267d578ba15e5c076fbdd8b2cb23355c5c3afbc6e108c26446bbf9b374","signature":"445df42ced0e4bafb8168c624dbf0fccd6aec3fec6f7bdb7c5533993c8e69190307abaeaf0583e669a998b34da7bf8f97ee2e2e89489d045f4bdb02687803209","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN1.acme/anchors","initiator":"612f2c9a306dc213e6f81781af1f1bb7b16b3d1d50f23ea1230663a22cbae03c"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN1.acme","sequenceNumber":1,"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"633ef8267d578ba15e5c076fbdd8b2cb23355c5c3afbc6e108c26446bbf9b374","signature":"ffac00877863132a209d988ae70d99590f5907bc99bf362acfbdfdbd46082ca3589969e9665a6c9178001c85461b2ae4b3358d4031fa1ad370d57b0a09c85f0a","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN2.acme/anchors","initiator":"853eb946ff51dc1dfb92b6dab1215e84c551da1bd465e3e82d2b38a51989e81c"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN2.acme","sequenceNumber":1,"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"633ef8267d578ba15e5c076fbdd8b2cb23355c5c3afbc6e108c26446bbf9b374","signature":"7174b6ff3ce6d82d1ec59ddb949a7e58e66ad7175b2ac1909e348dfe5797bf69a39ca671fe8061cfbef7f496572db6fd23bbd9ecd1c4e09cd6545c62fa41480e","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"1e17fd8766c93486299998719ff9bf0f76474e1acf729b4a32cb19f4d960c84a"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":1,"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"633ef8267d578ba15e5c076fbdd8b2cb23355c5c3afbc6e108c26446bbf9b374","signature":"9c62526605e3772525f47f3a18a8e2b93d8494571dc89c8f63261baf6f703b031b8f3e3862860fd90ff5d7b999ff02d99173ab5f1dfcb14fb68e85360722370d","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN2.acme/anchors","initiator":"853eb946ff51dc1dfb92b6dab1215e84c551da1bd465e3e82d2b38a51989e81c"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN2.acme","sequenceNumber":1,"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"9de8f58234a823c9b9618d852accb42559c0b16ae225f9bdbe9993731a4b6d5c","signature":"26c81167a3ecc6811f816c5831d4733b4ded4849af245dc003815c3b68c470c0421b697897ac34008d4d773167310ba0595f37d5a904e70fd76517d06adcd402","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"1e17fd8766c93486299998719ff9bf0f76474e1acf729b4a32cb19f4d960c84a"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":1,"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"9de8f58234a823c9b9618d852accb42559c0b16ae225f9bdbe9993731a4b6d5c","signature":"18de7bf744943bae61918f73defd2b52820ee0b3ab0bac95b64d8d6fd1155b8818f5076169276adc10ac564b8ed0c03f9796be2da91156048b5051e4e4b0ad0a","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN0.acme/anchors","initiator":"d8b499a53e6a0080678dd2f18b0f9fecd587860c8c797df0b8730ff4bc0d8574"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN0.acme","sequenceNumber":1,"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"9de8f58234a823c9b9618d852accb42559c0b16ae225f9bdbe9993731a4b6d5c","signature":"87eb899d616cefc81169ada22703fe2d2877b1147b0b0a75adc6c03e853b9ce94ad3d703bba9af1506ad0fbcae14431f4f1cf18d4e4a8b748102965926b28801","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN1.acme/anchors","initiator":"612f2c9a306dc213e6f81781af1f1bb7b16b3d1d50f23ea1230663a22cbae03c"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN1.acme","sequenceNumber":1,"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"9de8f58234a823c9b9618d852accb42559c0b16ae225f9bdbe9993731a4b6d5c","signature":"893859de7b7cc169c74d80af2f4d1481a58c64f1199826f30a2b222d4d4aa796244cf65d46f1564411f29f3c514f640e41a3c4de6081a583ea74b14e7935750c","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN2.acme/anchors","initiator":"853eb946ff51dc1dfb92b6dab1215e84c551da1bd465e3e82d2b38a51989e81c"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN2.acme","sequenceNumber":1,"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"78641be5bc680f8247265bf2da3cd31afd111c3f59379d80f095986d5fedd07a","signature":"3f81e8b4b324d80fd7a4002b8231e848f399a227278f4e6e8dd633a34f3fefc41a481a49c85dca9df1d23db556ee5f48e2da9f560cc1584e369cf338b8b8c50a","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"1e17fd8766c93486299998719ff9bf0f76474e1acf729b4a32cb19f4d960c84a"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":1,"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"78641be5bc680f8247265bf2da3cd31afd111c3f59379d80f095986d5fedd07a","signature":"828c4320c92ba488d02253afe6dc1ba2630eabeb81719dea0a0c632fc026292d10d91f8008b4cb54756b9ea2103bbdc6412e56d7625058fce8fc740deb06f10a","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN0.acme/anchors","initiator":"d8b499a53e6a0080678dd2f18b0f9fecd587860c8c797df0b8730ff4bc0d8574"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN0.acme","sequenceNumber":1,"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"78641be5bc680f8247265bf2da3cd31afd111c3f59379d80f095986d5fedd07a","signature":"dc25ee37295c002c7ade77aac69101fa5274d93fbab2b84f44b796d6bc1ece10e1720eb5a04f0c025f710e1af9b3c4e51e10be660792ee4cc12895f4a22bf80b","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN1.acme/anchors","initiator":"612f2c9a306dc213e6f81781af1f1bb7b16b3d1d50f23ea1230663a22cbae03c"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN1.acme","sequenceNumber":1,"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"78641be5bc680f8247265bf2da3cd31afd111c3f59379d80f095986d5fedd07a","signature":"9425b3bf4b7a5d45a79b84783afbfb3060f6e7ef35c4e3602b6b455a82b0056abae6181d81a71a7307f82eb4a682e7d05ab7ecd7108143b2ecead0d2c3c12b05","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN0.acme/anchors","initiator":"d8b499a53e6a0080678dd2f18b0f9fecd587860c8c797df0b8730ff4bc0d8574"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN0.acme","sequenceNumber":1,"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"05f3fcc7e95c17eaab6692d673176249f114a16bc28d2996eb5c08f6c6880320","signature":"fa6ad150d6ecd3c3d40481491bbfdcc3e8f5eef7d7c1633e19f68cdcf6e58396936ed0769896c3af09c35686a3dbfb1a1ea7627fe6681242667f0ac931337105","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},"transactionHash":"8a1a2c7bbb30f696d665ea952f34c256629ec6e680439d9ad28e160d56f4d3ac"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN1.acme/anchors","initiator":"612f2c9a306dc213e6f81781af1f1bb7b16b3d1d50f23ea1230663a22cbae03c"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN1.acme","sequenceNumber":1,"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"05f3fcc7e95c17eaab6692d673176249f114a16bc28d2996eb5c08f6c6880320","signature":"64b38a22965732a6982e9a3299299db0c6ea4cce15a8a4361c038d2c1807ebcf3219f1bb4ae0c0469e585f1cdfe4c826fafc23fb0178c302c81744a4d098bb0a","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},"transactionHash":"cc3c4848cb91fbd5be3245f21ea2044f0eb64291e7d09c0f3b62897b615a108d"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN2.acme/anchors","initiator":"853eb946ff51dc1dfb92b6dab1215e84c551da1bd465e3e82d2b38a51989e81c"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN2.acme","sequenceNumber":1,"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"05f3fcc7e95c17eaab6692d673176249f114a16bc28d2996eb5c08f6c6880320","signature":"f505f19da7b7bb557888a867caf6650d20349974052c80bdf18ee33d307100c9b4d57aab42b1c0a170f63b0478205f5ea77f89712ca656e9f6e416f52fe1cf0a","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},"transactionHash":"95579b49738102ce1daf8a51afe56ccdc70191dc702051bc60a39d84ba48f435"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"1e17fd8766c93486299998719ff9bf0f76474e1acf729b4a32cb19f4d960c84a"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c","makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":1,"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"05f3fcc7e95c17eaab6692d673176249f114a16bc28d2996eb5c08f6c6880320","signature":"a8fb3555525df84c29e9c6afed652d7828eb87973e445043a39d2cc8629845541458fff66c195afc8cab34f7e60550fa78ee2dde56a634ecb90464c861f6cd09","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},"transactionHash":"498a36fd1d008799703919a1c10e595a5743f8b6f9f38d2c14b951d4272ed7eb"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"bd358d14fed5510982ea97cb400471d333a198aeae24c1c0f6e8b59b3b651250"},"body":{"type":"blockValidatorAnchor","source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368","acmeBurnt":"0"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://bvn-BVN0.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":1,"transactionHash":"82f5d14dcb442897a6951843357fa0c74db83be513a19d9cd032e5966a644253"},"transactionHash":"82f5d14dcb442897a6951843357fa0c74db83be513a19d9cd032e5966a644253"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"b47dc38466953fa79e81abfc7bd19d564f870ec07f26aa1fb06751ea4bf7c4b1","signature":"3324d67787d11bcc3333eb539db3421c39b8f53a98bef7535eca02c5e9540f566a1af3c1d4692bcde8313dd0327cd8a5c357f06791276b4f885fae2c345a250a","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"82f5d14dcb442897a6951843357fa0c74db83be513a19d9cd032e5966a644253"},"transactionHash":"82f5d14dcb442897a6951843357fa0c74db83be513a19d9cd032e5966a644253"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"bd358d14fed5510982ea97cb400471d333a198aeae24c1c0f6e8b59b3b651250"},"body":{"type":"blockValidatorAnchor","source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368","acmeBurnt":"0"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://bvn-BVN0.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":1,"transactionHash":"82f5d14dcb442897a6951843357fa0c74db83be513a19d9cd032e5966a644253"},"transactionHash":"82f5d14dcb442897a6951843357fa0c74db83be513a19d9cd032e5966a644253"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"f0ba4305a861d7a3c1869f2632886e07134a03115f182f860b1113afd0121438","signature":"2fcfd867113907ce3c23a62a9ae9a9937a4f4554e66b2338351c1d8a2442bec4886405e8c09536aba061c04977089dc440f2c8fc5c87d9fad85e56978d734402","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"82f5d14dcb442897a6951843357fa0c74db83be513a19d9cd032e5966a644253"},"transactionHash":"82f5d14dcb442897a6951843357fa0c74db83be513a19d9cd032e5966a644253"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"bd358d14fed5510982ea97cb400471d333a198aeae24c1c0f6e8b59b3b651250"},"body":{"type":"blockValidatorAnchor","source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368","acmeBurnt":"0"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://bvn-BVN0.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":1,"transactionHash":"82f5d14dcb442897a6951843357fa0c74db83be513a19d9cd032e5966a644253"},"transactionHash":"82f5d14dcb442897a6951843357fa0c74db83be513a19d9cd032e5966a644253"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"60e7be68143da64401f57be2fe88640d06c0d314fea381ec7dc5a21e7516d758","signature":"2b6bf98ec4e15c5e0c8de20b1915049ada2d839d2e4bdafb42e00724547a879fadb656537e7fa27efcbb00f8bcac3b2c409177ab56973282158ec31693a6a60f","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"82f5d14dcb442897a6951843357fa0c74db83be513a19d9cd032e5966a644253"},"transactionHash":"82f5d14dcb442897a6951843357fa0c74db83be513a19d9cd032e5966a644253"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"7279014f3c331ab7ba0706c803d6198c707df7546ddba985c9c9b92378e683ab"},"body":{"type":"blockValidatorAnchor","source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138","acmeBurnt":"0"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://bvn-BVN1.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":1,"transactionHash":"2e246ecd727836f61b3f75f259a19620a5f106a88b11c3dd7e55708984aff4af"},"transactionHash":"2e246ecd727836f61b3f75f259a19620a5f106a88b11c3dd7e55708984aff4af"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"59d55ae3c2c644bf8c849d451eb8733b47f219850157decf2b7f3a5cd3c96c6b","signature":"163d01061295aef3c91b3b0f3a0fe3503f250ea6c6a946085d9be0f8068169a6e979b378cebf90d3ff6c114461a23ee1949b51f38e6f0122c3c81057471ff705","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"2e246ecd727836f61b3f75f259a19620a5f106a88b11c3dd7e55708984aff4af"},"transactionHash":"2e246ecd727836f61b3f75f259a19620a5f106a88b11c3dd7e55708984aff4af"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"7279014f3c331ab7ba0706c803d6198c707df7546ddba985c9c9b92378e683ab"},"body":{"type":"blockValidatorAnchor","source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138","acmeBurnt":"0"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://bvn-BVN1.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":1,"transactionHash":"2e246ecd727836f61b3f75f259a19620a5f106a88b11c3dd7e55708984aff4af"},"transactionHash":"2e246ecd727836f61b3f75f259a19620a5f106a88b11c3dd7e55708984aff4af"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"1a7551ffe00adda17a6c1af8f602dbdb4bdd7cbdc96f33cab1e938a2756f0e5d","signature":"8f9d4fc562da5656dbd987e47584e924610fc5f5b84b913f18acef8fce01dc4f7711eed05e3ee1719119f0153d387cba0c71ecf3b5a24b134dd6792031314607","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"2e246ecd727836f61b3f75f259a19620a5f106a88b11c3dd7e55708984aff4af"},"transactionHash":"2e246ecd727836f61b3f75f259a19620a5f106a88b11c3dd7e55708984aff4af"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"7279014f3c331ab7ba0706c803d6198c707df7546ddba985c9c9b92378e683ab"},"body":{"type":"blockValidatorAnchor","source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138","acmeBurnt":"0"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://bvn-BVN1.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":1,"transactionHash":"2e246ecd727836f61b3f75f259a19620a5f106a88b11c3dd7e55708984aff4af"},"transactionHash":"2e246ecd727836f61b3f75f259a19620a5f106a88b11c3dd7e55708984aff4af"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"633ef8267d578ba15e5c076fbdd8b2cb23355c5c3afbc6e108c26446bbf9b374","signature":"5c85395aa5310d8f285919fe14a6680ef256322e57d47f8c4cffb01d82a09a05752136ce7a0acfcc4de7e472c533c4ff5815cd9590c0c0435b881011032f9809","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"2e246ecd727836f61b3f75f259a19620a5f106a88b11c3dd7e55708984aff4af"},"transactionHash":"2e246ecd727836f61b3f75f259a19620a5f106a88b11c3dd7e55708984aff4af"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"0395f50a9e2c529339bbd1e03b8f83448233936aeb79d31fa03a89324a9c29ac"},"body":{"type":"blockValidatorAnchor","source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb","acmeBurnt":"0"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://bvn-BVN0.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":2,"transactionHash":"38e15e88842ebe008662af0076db366c8d56992c24a0ec44cde66283b3dfedcd"},"transactionHash":"38e15e88842ebe008662af0076db366c8d56992c24a0ec44cde66283b3dfedcd"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"b47dc38466953fa79e81abfc7bd19d564f870ec07f26aa1fb06751ea4bf7c4b1","signature":"3850dfb86a7e4227c467349b3308629f20f813162f25be712048f9f1dd6a325162de92cb472dd4894f91e640ccdd4f01ebf85d4145d749adfcffa62444cfe507","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"38e15e88842ebe008662af0076db366c8d56992c24a0ec44cde66283b3dfedcd"},"transactionHash":"38e15e88842ebe008662af0076db366c8d56992c24a0ec44cde66283b3dfedcd"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"0395f50a9e2c529339bbd1e03b8f83448233936aeb79d31fa03a89324a9c29ac"},"body":{"type":"blockValidatorAnchor","source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb","acmeBurnt":"0"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://bvn-BVN0.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":2,"transactionHash":"38e15e88842ebe008662af0076db366c8d56992c24a0ec44cde66283b3dfedcd"},"transactionHash":"38e15e88842ebe008662af0076db366c8d56992c24a0ec44cde66283b3dfedcd"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"f0ba4305a861d7a3c1869f2632886e07134a03115f182f860b1113afd0121438","signature":"0d4ef443f9e5e47bd253ba4e7899b4793886d2cd4f68852125c3367d92c80062f514cd87ccd2a693a8808ed349f9a6bd341962836224f6514d8b5b48d399e409","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"38e15e88842ebe008662af0076db366c8d56992c24a0ec44cde66283b3dfedcd"},"transactionHash":"38e15e88842ebe008662af0076db366c8d56992c24a0ec44cde66283b3dfedcd"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"0395f50a9e2c529339bbd1e03b8f83448233936aeb79d31fa03a89324a9c29ac"},"body":{"type":"blockValidatorAnchor","source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb","acmeBurnt":"0"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://bvn-BVN0.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":2,"transactionHash":"38e15e88842ebe008662af0076db366c8d56992c24a0ec44cde66283b3dfedcd"},"transactionHash":"38e15e88842ebe008662af0076db366c8d56992c24a0ec44cde66283b3dfedcd"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"60e7be68143da64401f57be2fe88640d06c0d314fea381ec7dc5a21e7516d758","signature":"a1f7841b57cbaa846cdbe3a1552b40fb8fbeecc39f643939e827db9e36e131655df1cfd575f14da699da4db302d3c0c5eed7e90ea6ff10d52e1868b0bf8c4908","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"38e15e88842ebe008662af0076db366c8d56992c24a0ec44cde66283b3dfedcd"},"transactionHash":"38e15e88842ebe008662af0076db366c8d56992c24a0ec44cde66283b3dfedcd"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN0.acme/anchors","initiator":"4a41165dc0ce0094436a0797299a2b4b48723049641f19b5772e396f478f191d"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN0.acme","sequenceNumber":2,"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"b47dc38466953fa79e81abfc7bd19d564f870ec07f26aa1fb06751ea4bf7c4b1","signature":"51f1fc57c8f416b0ed2e144a23a776ac5a4a3f8e9254604cf6a49c86f916607b9156532e857c951b9c492f8adcdf128146efadc4fc985c69b38ad1e35c43f101","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN1.acme/anchors","initiator":"c6706e7df770248428dbaf97e7624d6beec83541875b5300459603b534ec650f"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN1.acme","sequenceNumber":2,"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"b47dc38466953fa79e81abfc7bd19d564f870ec07f26aa1fb06751ea4bf7c4b1","signature":"3e6b74bd3e48f1914d140c7ef7f66e29472f9c59ff29b38d9bde1e53c26ec044c42dba10f9d6a5946246aaaabe66b66e2ffb132a8227e46e312e92a791ccf80d","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN2.acme/anchors","initiator":"6fdcd92fefb3bac012db6c4d8702e168ada2b379e210e6b943ee13c66891b424"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN2.acme","sequenceNumber":2,"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"b47dc38466953fa79e81abfc7bd19d564f870ec07f26aa1fb06751ea4bf7c4b1","signature":"15a07fdb28531978b5ada74717bf597eb26a41659964352d62ef0456f98a57b7dade632b0b733a5b7e833538adc368c4338f5b9c1d0ee60d40c1e546dceb7c08","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"59f9744884c068135e9c4aa04249c0301f1e4e421f98038ded9a6c3634b31939"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":2,"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"b47dc38466953fa79e81abfc7bd19d564f870ec07f26aa1fb06751ea4bf7c4b1","signature":"8f6bccc8187ebda74f37ce5b9ac977235318a867c265edd7db2f1e36a371410e274a04cddd57ec5e5d31ad40f39009ba2a80f45e66674cf3a7e09de648a37d00","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN0.acme/anchors","initiator":"4a41165dc0ce0094436a0797299a2b4b48723049641f19b5772e396f478f191d"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN0.acme","sequenceNumber":2,"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"f0ba4305a861d7a3c1869f2632886e07134a03115f182f860b1113afd0121438","signature":"f30640f90a7ff6c5bf69337cf0569abb7be33562a7077940dd38e26a2bfe8961c4fb7f7683837ae96977bf78c89bceb14d47fc80f1ede5ea896e533592e75b0f","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN0.acme/anchors","initiator":"4a41165dc0ce0094436a0797299a2b4b48723049641f19b5772e396f478f191d"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN0.acme","sequenceNumber":2,"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"60e7be68143da64401f57be2fe88640d06c0d314fea381ec7dc5a21e7516d758","signature":"223da8f7718a35dd4841d8164d48847a53c6fa0b455ccce4c5e171a4928b816334d05d7e5b5af451892358a55953507ded2125792f9d0000eadcbff00a846a07","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN0.acme/anchors","initiator":"4a41165dc0ce0094436a0797299a2b4b48723049641f19b5772e396f478f191d"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN0.acme","sequenceNumber":2,"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"78641be5bc680f8247265bf2da3cd31afd111c3f59379d80f095986d5fedd07a","signature":"e49c79de79f2a20d8f922670314c3169cc94ef993e0c43b8ab978eb445128c26094e0e535c55dd5d64924982107177b690099d29763de57c38e70dc2baf1410a","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN1.acme/anchors","initiator":"c6706e7df770248428dbaf97e7624d6beec83541875b5300459603b534ec650f"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN1.acme","sequenceNumber":2,"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"59d55ae3c2c644bf8c849d451eb8733b47f219850157decf2b7f3a5cd3c96c6b","signature":"80a4efd454113df52e13770db4871142bb3c23a15ff969ddcd432a325fcbea743b68a7e0bfcc9f7329e4ba3ab4a9e7593bb6976e5f59daaf471b3ba234fb9808","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN1.acme/anchors","initiator":"c6706e7df770248428dbaf97e7624d6beec83541875b5300459603b534ec650f"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN1.acme","sequenceNumber":2,"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"f0ba4305a861d7a3c1869f2632886e07134a03115f182f860b1113afd0121438","signature":"b697ce91ffc86f83e330128ddc83bb3ef2543569a17ca71e8465d1d92fbadf8cf43504ab06cb71bd8ea6f36ced97a0d2bf6f2bf1c1a05f1c62565d9d35a5fa02","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN1.acme/anchors","initiator":"c6706e7df770248428dbaf97e7624d6beec83541875b5300459603b534ec650f"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN1.acme","sequenceNumber":2,"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"60e7be68143da64401f57be2fe88640d06c0d314fea381ec7dc5a21e7516d758","signature":"8e16eaaccd7c030e1828be948d2d09c4b7546d3fe9ea5affcd4de0cc8f49e74b0d6a53e2085c209417dc74b7e5e200a6966862699dee1fd14fbc31e4f9ce4201","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN1.acme/anchors","initiator":"c6706e7df770248428dbaf97e7624d6beec83541875b5300459603b534ec650f"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN1.acme","sequenceNumber":2,"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"78641be5bc680f8247265bf2da3cd31afd111c3f59379d80f095986d5fedd07a","signature":"8e0cc79fafd56a1915823dc523b9dd59f1265b92ffc30b6d14c51e058df95ef0f3c5e5ea3c2e38e13fd105fd18f5f504a0a510812a474a0149911288b7340704","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN2.acme/anchors","initiator":"6fdcd92fefb3bac012db6c4d8702e168ada2b379e210e6b943ee13c66891b424"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN2.acme","sequenceNumber":2,"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"633ef8267d578ba15e5c076fbdd8b2cb23355c5c3afbc6e108c26446bbf9b374","signature":"b612bddd5d1664247fdbea54b2bb847dbd66d464109903d388ed89649f4b77b5556456aa5fe9b308e83d19fa29810dda426856a8cc928ca1f38c3177795df301","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"59f9744884c068135e9c4aa04249c0301f1e4e421f98038ded9a6c3634b31939"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":2,"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"633ef8267d578ba15e5c076fbdd8b2cb23355c5c3afbc6e108c26446bbf9b374","signature":"1e6f5aadfd3a72200d660d481416237381f1ed5275cfebee3e8ba20294f636478bf41faa008246d8d137b3abd8ece1f243dab2553ff66cb39fadcfa5f2735004","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN0.acme/anchors","initiator":"4a41165dc0ce0094436a0797299a2b4b48723049641f19b5772e396f478f191d"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN0.acme","sequenceNumber":2,"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"633ef8267d578ba15e5c076fbdd8b2cb23355c5c3afbc6e108c26446bbf9b374","signature":"3f76d86442bd3b8bfa74dc595030bebe1b2393b6604204336dd34d973c726ed5ee89d9032e8ea8be8d87601001cdc58e86d63dca1085f6469c205eecce27f804","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN1.acme/anchors","initiator":"c6706e7df770248428dbaf97e7624d6beec83541875b5300459603b534ec650f"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN1.acme","sequenceNumber":2,"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"633ef8267d578ba15e5c076fbdd8b2cb23355c5c3afbc6e108c26446bbf9b374","signature":"4bbfea73a12e10c937c7a7f1953ed96d7bd236cabe0ff30c94c09234280ad91a53ce7568b9b666603cb57fd6c9abda86394642f56fb06e5f01529f0f8fd7c006","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"59f9744884c068135e9c4aa04249c0301f1e4e421f98038ded9a6c3634b31939"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":2,"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"1a7551ffe00adda17a6c1af8f602dbdb4bdd7cbdc96f33cab1e938a2756f0e5d","signature":"3fbe39cec95dbfdbb1eba1f32aff1060a930902a8c7a0d5b8e19ff5a4491c349b9fb7ff68e8ba54b414f90d6e216f26c04d07c46031f99f6715fa70b4744d70a","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN0.acme/anchors","initiator":"4a41165dc0ce0094436a0797299a2b4b48723049641f19b5772e396f478f191d"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN0.acme","sequenceNumber":2,"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"1a7551ffe00adda17a6c1af8f602dbdb4bdd7cbdc96f33cab1e938a2756f0e5d","signature":"0a2beabf1dbdf281aa445d0846caccb66651794123121be6268020bebf4f058f2e7ce2d06efd62873640633032b32dfdf1df6e4f4cef3d79a049d11c6b413d01","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN1.acme/anchors","initiator":"c6706e7df770248428dbaf97e7624d6beec83541875b5300459603b534ec650f"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN1.acme","sequenceNumber":2,"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"1a7551ffe00adda17a6c1af8f602dbdb4bdd7cbdc96f33cab1e938a2756f0e5d","signature":"2794d7158020adbb5739f0f5bb41c001340e843ecce973f23774e3460e57d4b0763b4dfc4c0ef0ab2b8ade27bf7009d42605d7bef94c62a38b472604d65ab90c","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN2.acme/anchors","initiator":"6fdcd92fefb3bac012db6c4d8702e168ada2b379e210e6b943ee13c66891b424"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN2.acme","sequenceNumber":2,"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"1a7551ffe00adda17a6c1af8f602dbdb4bdd7cbdc96f33cab1e938a2756f0e5d","signature":"e61b9e2fd661f3d54cebd9c6ee745480d4928e8a928873d3c575a7e82bcde906721f8b8d8aa73eb39f1e5a0e91ef8e1b997d04cee444a0572c993082cbda9008","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN2.acme/anchors","initiator":"6fdcd92fefb3bac012db6c4d8702e168ada2b379e210e6b943ee13c66891b424"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN2.acme","sequenceNumber":2,"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"05f3fcc7e95c17eaab6692d673176249f114a16bc28d2996eb5c08f6c6880320","signature":"c800a2d59f10b83ca075ecce410eb8ba6b8284334b65676e224c6b8aad7d7346fdaad46526472b36bb58c23b8608c801a088ebe6c289afc84e720a468bf89d06","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"59f9744884c068135e9c4aa04249c0301f1e4e421f98038ded9a6c3634b31939"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":2,"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"05f3fcc7e95c17eaab6692d673176249f114a16bc28d2996eb5c08f6c6880320","signature":"9ae413bb3567f11effc4507bf3a02efbe75c5ffecb781b24b0087eb35f2a65adbbaf8e7b7318633fd3ed480105a53891782f899bb13b7e06ab28e7870ce6f00f","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN0.acme/anchors","initiator":"4a41165dc0ce0094436a0797299a2b4b48723049641f19b5772e396f478f191d"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN0.acme","sequenceNumber":2,"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"05f3fcc7e95c17eaab6692d673176249f114a16bc28d2996eb5c08f6c6880320","signature":"7ae377cb5354c06e541a7a92733b9747f879d7e0b520e46b604bfbdea6bc326ff3be225bdd0b78bd83cdf6195505d11cfd2366625d122140c22c366162cca40f","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN1.acme/anchors","initiator":"c6706e7df770248428dbaf97e7624d6beec83541875b5300459603b534ec650f"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN1.acme","sequenceNumber":2,"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"05f3fcc7e95c17eaab6692d673176249f114a16bc28d2996eb5c08f6c6880320","signature":"495cb2732fcdd9a8e63ce253364ff6e28639fbadbd03a20b3ff5f66f2477ebcc1bec0493bc6a8b133b869970a45225edc6ce2e56e5751b3d664825280b5b670f","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"59f9744884c068135e9c4aa04249c0301f1e4e421f98038ded9a6c3634b31939"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":2,"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"9de8f58234a823c9b9618d852accb42559c0b16ae225f9bdbe9993731a4b6d5c","signature":"c87ccd38acc9b109d00af2beb215cd509309732a5a31d83f73b92a0c8a65646d72ccc555dcbb3f0b644904dd373d92d2c2dae9ededa35ccf79b54c0fc92d9b0f","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN0.acme/anchors","initiator":"4a41165dc0ce0094436a0797299a2b4b48723049641f19b5772e396f478f191d"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN0.acme","sequenceNumber":2,"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"9de8f58234a823c9b9618d852accb42559c0b16ae225f9bdbe9993731a4b6d5c","signature":"2e102d6a145569ff1f0e7a089cfe0cce4c74cc2a4d679862c569b6a68c020c6b067f130bea45d461bae127e1ce82744d975aa410d410101a13e5dfae8b99710d","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN1.acme/anchors","initiator":"c6706e7df770248428dbaf97e7624d6beec83541875b5300459603b534ec650f"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN1.acme","sequenceNumber":2,"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"9de8f58234a823c9b9618d852accb42559c0b16ae225f9bdbe9993731a4b6d5c","signature":"311a2226322714ae88c23fbfbe414201abe7e346685d28763b5de1882a251eb9a97bba1be63c3c9e333ba425f8a2af31dd963330ad533c6cc11a2bb4504ef408","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},"transactionHash":"0a483b66fcd51da5aae093d3293a5d94528d35aa9aeb1e5cc408a20f68a587a3"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN2.acme/anchors","initiator":"6fdcd92fefb3bac012db6c4d8702e168ada2b379e210e6b943ee13c66891b424"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN2.acme","sequenceNumber":2,"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"9de8f58234a823c9b9618d852accb42559c0b16ae225f9bdbe9993731a4b6d5c","signature":"4704183bfaabbfbec0c232c19e9f34200c8967c8106aaf1785fac9b9998ba7b6de1595df5cd64e72d7de0dfc9e4c540c4a96d242928d238156b8847bd5543803","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN2.acme/anchors","initiator":"6fdcd92fefb3bac012db6c4d8702e168ada2b379e210e6b943ee13c66891b424"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN2.acme","sequenceNumber":2,"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"59d55ae3c2c644bf8c849d451eb8733b47f219850157decf2b7f3a5cd3c96c6b","signature":"1f18a1c01dc2e538f8c24e528a1bbb319f31065a7a4ba44cd11fec58af0bbea2bb806786bee7d84d5d686c5b1215345dc003c57bea84042fb32a204db5e05e02","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"59f9744884c068135e9c4aa04249c0301f1e4e421f98038ded9a6c3634b31939"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":2,"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"59d55ae3c2c644bf8c849d451eb8733b47f219850157decf2b7f3a5cd3c96c6b","signature":"c577e259767e4c89617ec30bbf76c1d1653e8e8d68ca7f0fdf8e39b07911ae472856f7d7475a85c9d55f46d23d403d1a0f4be165cc8d8c21fe75ae293f6ba701","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN0.acme/anchors","initiator":"4a41165dc0ce0094436a0797299a2b4b48723049641f19b5772e396f478f191d"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN0.acme","sequenceNumber":2,"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"59d55ae3c2c644bf8c849d451eb8733b47f219850157decf2b7f3a5cd3c96c6b","signature":"1dcbaca8e4208dd9aab900b1dce1ed26fe3ed0265220f343b15f4708e083d183921f42d967da988fbe1fa8a67eaa9ea685fd841c8e27eee783b1556ecfa21a09","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},"transactionHash":"6630b9dd670cdac3b8712c46c7f533aa529dfb63072beb4dadf38a30d7aab751"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN2.acme/anchors","initiator":"6fdcd92fefb3bac012db6c4d8702e168ada2b379e210e6b943ee13c66891b424"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN2.acme","sequenceNumber":2,"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"f0ba4305a861d7a3c1869f2632886e07134a03115f182f860b1113afd0121438","signature":"7a482740c98cc890668c802300848810c81e3690b1fd120b361d93c65fde7fd68c4b082ee4ad0aa75833c36c8e0635584c374abe10d11129a5d9bd0ee3b0750a","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"59f9744884c068135e9c4aa04249c0301f1e4e421f98038ded9a6c3634b31939"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":2,"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"f0ba4305a861d7a3c1869f2632886e07134a03115f182f860b1113afd0121438","signature":"0f1530b933239d493150f2abe7dcb77cfcdfa4d0d73436cfff5d67b917e3f826b64bdeefb15b0b3894b01a8a932a48f6ee3c3f49ace04b0a10ce142017347107","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN2.acme/anchors","initiator":"6fdcd92fefb3bac012db6c4d8702e168ada2b379e210e6b943ee13c66891b424"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN2.acme","sequenceNumber":2,"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"60e7be68143da64401f57be2fe88640d06c0d314fea381ec7dc5a21e7516d758","signature":"11f59d02411f138dd72eff26b7373a6e02baf070674d37d89144b7a18d6d7f277c02c5fe0c4bbfca5a8f7b75c296502ee4df2172cc2c18b2ed28c52c8473c90e","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"59f9744884c068135e9c4aa04249c0301f1e4e421f98038ded9a6c3634b31939"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":2,"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"60e7be68143da64401f57be2fe88640d06c0d314fea381ec7dc5a21e7516d758","signature":"dd13e3941beb728ffb856974a9251ce73b474079f91b36083f19b81bc9dad13512d13e6894fce861983d7d4d6ef5c532c1194d1f6a01d4d431f8b8ae7422320b","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN2.acme/anchors","initiator":"6fdcd92fefb3bac012db6c4d8702e168ada2b379e210e6b943ee13c66891b424"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN2.acme","sequenceNumber":2,"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"78641be5bc680f8247265bf2da3cd31afd111c3f59379d80f095986d5fedd07a","signature":"9f693356bfad318c4eb74f96e61c1a1510386026c6c261db006adff8e50b57cc391031cae5563d939e88d8b12ba3c03675bd1dda0e25bfc9acbb3f0d7756250a","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},"transactionHash":"a58c365c40d85f0af099a4130b82db2873e761c90eaff58f9280061802fff92e"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"59f9744884c068135e9c4aa04249c0301f1e4e421f98038ded9a6c3634b31939"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":4,"rootChainIndex":22,"rootChainAnchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","stateTreeAnchor":"67b62f3ebd73eea8615af8668737e9e1d92420d8f69dd495b0a12645acd257d7","receipts":[{"anchor":{"source":"acc://bvn-BVN1.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","stateTreeAnchor":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},"rootChainReceipt":{"start":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","end":"39a1d82b40fb0245c19016e54965da9a454c02dff98d8512953eb72f522e3280","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"0c60991bd0e2350adcaf375e3d4387bebcbb75dfb3baddf4d7e63b03695c4138"},{"right":true,"hash":"9bc4bb4f9d621ee8dadeb2baa4fef5040ee76b610881cc8f47acdaaaa8b8f65f"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://dn.acme","minorBlockIndex":1,"rootChainIndex":10,"rootChainAnchor":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","stateTreeAnchor":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},"rootChainReceipt":{"start":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","end":"574b9e3a1de84d0b0faa15d25546172f225a82011196e43214426f040fbb2803","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a2b28ecc1d6e1ec06fb2e5a7bee3a4158d53a0a81974ac1e3aef89af012d0d1c"},{"right":true,"hash":"96e9afda74132177cccca6dbffb4ecafaf562ad61a81b0278de62b0a4695a9bf"},{"hash":"b115f0ee78635f572594f2d108b8648aaadef5352502abac2fb1cd2ad2a6529a"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}},{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":1,"rootChainIndex":11,"rootChainAnchor":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","stateTreeAnchor":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},"rootChainReceipt":{"start":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","end":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"cb698fc1b4fbae7f4594c5dd91b2221e2c0effb7990932835db24c86626db368"},{"hash":"2c64507c1a25ef418e60fa9115c17d608cb4bfae071723f7e24fd99d0d503b92"},{"hash":"1e33961628a410d51b248f45baa8cda2d4dd6fbc59200374d10e0ebea24265fc"},{"hash":"df5d15d7e8568a055ade13754b86d6a3eec61c77ec82b301aa616167efd499f1"},{"right":true,"hash":"5ad75d6147b58625423381f888376d9548bd8ab9ffd6f1372931988209b716b2"}]}},{"anchor":{"source":"acc://bvn-BVN2.acme","minorBlockIndex":1,"rootChainIndex":9,"rootChainAnchor":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","stateTreeAnchor":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},"rootChainReceipt":{"start":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","end":"55662afbaa2425fcced34d860f68b82a20db9879bbe309d28b8454378d661b9a","anchor":"bb19c088290b27759096c9bccab0f0056f5399b3c94a06f3c61ab3a2edacb639","entries":[{"hash":"a1186ddcea9f42131a61bd005d783cc5f6d2702b111b148fb7d23e0d81e04fa2"},{"hash":"63f1b2a2d7ad882abea1241448f555a54a3297f76f304ad03d5f0af7813aad97"},{"right":true,"hash":"17f6a2be5b57632edb9ce3d22989e35c9bb1dc68e0e7f8242101b0a203d9c309"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":2,"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"78641be5bc680f8247265bf2da3cd31afd111c3f59379d80f095986d5fedd07a","signature":"9caa2440b91be5eba23956bff0171d9952e703d3c706afe02d1d1b0d91c14b275517c0a66834024ff17c801777ab42daf24749af2e793773cfb17b7038a4ac0a","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},"transactionHash":"cca825a356047dc593e0c9c456f5bef5e4e1e99064c551540986dc4aa3014acf"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN0.acme/anchors","initiator":"f0a0e14476c34242df6a769de4725f3399ba84496c8dd9195dffced05c2ced7c"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN0.acme","sequenceNumber":3,"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"b47dc38466953fa79e81abfc7bd19d564f870ec07f26aa1fb06751ea4bf7c4b1","signature":"414f5d7b08df4e9ba2aaf2eca07232852fc3ef23ffca85ad5ee443ed34d248a964d41d23595a08020e72f0dcea613464a3127ef8e678d3d34e8cba0a6ee76c08","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN1.acme/anchors","initiator":"77d7dd4e8855a911b63b77d42c8379ecf1d792933ec20490c1665ee111f5b011"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN1.acme","sequenceNumber":3,"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"b47dc38466953fa79e81abfc7bd19d564f870ec07f26aa1fb06751ea4bf7c4b1","signature":"ab4c41adc0181d8485343a6a892c2595d9a2a7207c57768c39c2f5621b81d7bd2a80a4a99da0e80973374647fe45554ad56a1b21e7990bf8d415e6a9795d210d","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN2.acme/anchors","initiator":"4e85f0d164b8df8d6e21586d2e7c5ad11f09c337c32d6da386b420c5c89254df"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN2.acme","sequenceNumber":3,"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"b47dc38466953fa79e81abfc7bd19d564f870ec07f26aa1fb06751ea4bf7c4b1","signature":"294f5d82762a9fb89ff9c0ec98a1cc4cfa506960b9448e76f848dcaf81f9979cda97013ffeeaf8c9d93cfdf560b1b83f5cb0d48d0a96b0b7bce56516012b1104","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"3e13dd70dec6a74e369e383bb7763e780959297c3d96eca68e286e201fe2a6c9"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":3,"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"b47dc38466953fa79e81abfc7bd19d564f870ec07f26aa1fb06751ea4bf7c4b1","signature":"5f94cae2288b79ccf396abc6333658e0b09b2a191512c60ca62ef0ac19aeb6bd8792b23cc9d516e3748ae30f3a79a3df0deabcd991c6c1666582a7c7faa65302","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN2.acme/anchors","initiator":"4e85f0d164b8df8d6e21586d2e7c5ad11f09c337c32d6da386b420c5c89254df"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN2.acme","sequenceNumber":3,"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"f0ba4305a861d7a3c1869f2632886e07134a03115f182f860b1113afd0121438","signature":"fbda4eb5aa393bad56731ebba19d85874fddf657710d3dc04a60c6f5fec723fce236b5fcb554ea9acc8599c34267bc85b253d524755661287f9d486a54f3910d","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"3e13dd70dec6a74e369e383bb7763e780959297c3d96eca68e286e201fe2a6c9"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":3,"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"f0ba4305a861d7a3c1869f2632886e07134a03115f182f860b1113afd0121438","signature":"14a705d8e9db4c5a8147d4c659850f202c2b2c9c19fcb34fa9e407c97503ac046059707181c2a1aaa694b9153ccf66f13e3e52c0703b81d3b9f61e847eae3a05","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN0.acme/anchors","initiator":"f0a0e14476c34242df6a769de4725f3399ba84496c8dd9195dffced05c2ced7c"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN0.acme","sequenceNumber":3,"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"f0ba4305a861d7a3c1869f2632886e07134a03115f182f860b1113afd0121438","signature":"83042c5e3cb3eef8013843aed698e9ba609e64dff7837a7e47cb9281c9158476afbf1620733adb026646029247d2e9cbad63f48b81f8303b826bd34ce20fdf0b","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN1.acme/anchors","initiator":"77d7dd4e8855a911b63b77d42c8379ecf1d792933ec20490c1665ee111f5b011"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN1.acme","sequenceNumber":3,"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"f0ba4305a861d7a3c1869f2632886e07134a03115f182f860b1113afd0121438","signature":"184ca757c4e300e7fb4dfd955f543f1f8c6e4d57f2b1820cba296e630e511c8fefde915cf5f614b74211c44a0c48a1e9dd931d97f5fc5e924c6be8c6b56baf0c","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN2.acme/anchors","initiator":"4e85f0d164b8df8d6e21586d2e7c5ad11f09c337c32d6da386b420c5c89254df"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN2.acme","sequenceNumber":3,"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"60e7be68143da64401f57be2fe88640d06c0d314fea381ec7dc5a21e7516d758","signature":"9ebb0c259abb4badbed89b46f10eb9dfa963d9fba9763525e359538cbf5482830c19fafa7cef370086e3c4047528e8aa2d4c68561d638cdbe82c18a845073c0f","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"3e13dd70dec6a74e369e383bb7763e780959297c3d96eca68e286e201fe2a6c9"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":3,"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"60e7be68143da64401f57be2fe88640d06c0d314fea381ec7dc5a21e7516d758","signature":"262e37124c0b479dac1ee77853126bf04ecc44d45b61034bc120d1f2ce87f4d571863a234347a6911179d3bcfaa4f786ab2e980c266e74266fca87136431bd0a","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN0.acme/anchors","initiator":"f0a0e14476c34242df6a769de4725f3399ba84496c8dd9195dffced05c2ced7c"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN0.acme","sequenceNumber":3,"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"60e7be68143da64401f57be2fe88640d06c0d314fea381ec7dc5a21e7516d758","signature":"c4a68dfca1987e3d704c0c772eea171a81cacbecbb904695669b3f8fc5a6ed4745cf2ac55069547ce223c3cb07451c97f765c7faf65ec88b9545ca89b851b50a","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN1.acme/anchors","initiator":"77d7dd4e8855a911b63b77d42c8379ecf1d792933ec20490c1665ee111f5b011"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN1.acme","sequenceNumber":3,"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"60e7be68143da64401f57be2fe88640d06c0d314fea381ec7dc5a21e7516d758","signature":"36bbef327a52d044ca458161e70e366eb80309f421c263d77d1f66b1954d13cbb79b0f3a2954a94ae12e560175410c049043c711c9a9153e6ec94576a4ff650f","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN0.acme/anchors","initiator":"f0a0e14476c34242df6a769de4725f3399ba84496c8dd9195dffced05c2ced7c"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN0.acme","sequenceNumber":3,"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"59d55ae3c2c644bf8c849d451eb8733b47f219850157decf2b7f3a5cd3c96c6b","signature":"909fde8e19c48648bddcade79b7f8c92a50f54d411b2af43de8c401a1c2e02e3191c8340278539fbfac7a2089de9cacb02d57966c37d53808d6c07ada9b4560a","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN1.acme/anchors","initiator":"77d7dd4e8855a911b63b77d42c8379ecf1d792933ec20490c1665ee111f5b011"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN1.acme","sequenceNumber":3,"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"59d55ae3c2c644bf8c849d451eb8733b47f219850157decf2b7f3a5cd3c96c6b","signature":"f73b02d8e67fa8cae36ba2380a5e7becdf903f42d00f22645202a7d1eb6beee404c90e7b728c3852010dd92bd91de979d3de5fb13d6218f611108e24ebd8530d","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN2.acme/anchors","initiator":"4e85f0d164b8df8d6e21586d2e7c5ad11f09c337c32d6da386b420c5c89254df"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN2.acme","sequenceNumber":3,"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"59d55ae3c2c644bf8c849d451eb8733b47f219850157decf2b7f3a5cd3c96c6b","signature":"cc675fc029e012cd84b7b81f01d36f506e01eeb6cb3497502d55b9425af204c615e9ff200bafea0a392c41137aa094938515bc168aea58f068226852ff5c810c","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"3e13dd70dec6a74e369e383bb7763e780959297c3d96eca68e286e201fe2a6c9"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":3,"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"59d55ae3c2c644bf8c849d451eb8733b47f219850157decf2b7f3a5cd3c96c6b","signature":"efc60184adeee595ac5a0ec64980ff0887928cd09f4db71da3ef2fe6245f9dbe39ddfe4d0e85ed2937cf406b675798de5dc375d98e204d86c9f1ce06de7c4500","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN0.acme/anchors","initiator":"f0a0e14476c34242df6a769de4725f3399ba84496c8dd9195dffced05c2ced7c"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN0.acme","sequenceNumber":3,"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"1a7551ffe00adda17a6c1af8f602dbdb4bdd7cbdc96f33cab1e938a2756f0e5d","signature":"f96036dbcedec9564b78fb2bb0cd8e68fad255c28a6bd1aa5690e42f6224e68cc8bb86217b9cbb5d5156f8d1b28385e3188106f4059a0c2882d1697a3e639f05","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN1.acme/anchors","initiator":"77d7dd4e8855a911b63b77d42c8379ecf1d792933ec20490c1665ee111f5b011"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN1.acme","sequenceNumber":3,"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"1a7551ffe00adda17a6c1af8f602dbdb4bdd7cbdc96f33cab1e938a2756f0e5d","signature":"859b302b7e37d8e1477294da15a4f0717e21f7f80728ed573fc32761e470d30f97e5342bc816ca974a05683fd982633eee6f4dfdda6c538757a0a1dfc7d3980f","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN2.acme/anchors","initiator":"4e85f0d164b8df8d6e21586d2e7c5ad11f09c337c32d6da386b420c5c89254df"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN2.acme","sequenceNumber":3,"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"1a7551ffe00adda17a6c1af8f602dbdb4bdd7cbdc96f33cab1e938a2756f0e5d","signature":"d075e7a7ad9db6386932c899f9fcccf6f4f77615079eb6ea561c00630a654b760821078e056102817edcc44a5d6dfae6b6bfdfe33d948ec96dbf513e97916b08","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"3e13dd70dec6a74e369e383bb7763e780959297c3d96eca68e286e201fe2a6c9"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":3,"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"1a7551ffe00adda17a6c1af8f602dbdb4bdd7cbdc96f33cab1e938a2756f0e5d","signature":"c24984caebc3d80151bbedba63db8a884a3afb20d2abcef0f890b3052c156321dfad39948bfa330c5e9790ecafd6bb2ea98e1bb047b01b82b6dc19723f56b408","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN0.acme/anchors","initiator":"f0a0e14476c34242df6a769de4725f3399ba84496c8dd9195dffced05c2ced7c"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN0.acme","sequenceNumber":3,"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"633ef8267d578ba15e5c076fbdd8b2cb23355c5c3afbc6e108c26446bbf9b374","signature":"56604949bb9b637db74dac73c80661c920cd520c209598e9b330c217b15cc5edb7503d5714eb484bce79b27173aec296b238449e7833b704098c358a6660a80b","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN1.acme/anchors","initiator":"77d7dd4e8855a911b63b77d42c8379ecf1d792933ec20490c1665ee111f5b011"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN1.acme","sequenceNumber":3,"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"633ef8267d578ba15e5c076fbdd8b2cb23355c5c3afbc6e108c26446bbf9b374","signature":"05ab3ed13e768370a997330b2fd2d48df77744c445620183a826df03a7ee64f718f551b4d6b1d7961baf55d3e28384580eb6bd41c42eaff861024aa2a71bf304","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN2.acme/anchors","initiator":"4e85f0d164b8df8d6e21586d2e7c5ad11f09c337c32d6da386b420c5c89254df"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN2.acme","sequenceNumber":3,"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"633ef8267d578ba15e5c076fbdd8b2cb23355c5c3afbc6e108c26446bbf9b374","signature":"1b5791d6146f7e564000663ee3e2a2914abaaea091b65e488cad5a5d2eeea3bcecb4f748cff5482bd42956b9973b4477749a4ec23e706f8ac55a3d41bb3ce004","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"3e13dd70dec6a74e369e383bb7763e780959297c3d96eca68e286e201fe2a6c9"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":3,"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"633ef8267d578ba15e5c076fbdd8b2cb23355c5c3afbc6e108c26446bbf9b374","signature":"342c101325b93cb4c4dc80d618c3904e40f30c38850309a5851545ca30f8d29ed05d94cf880b557f1eeab0c699c248136d9ca26d53190f4d8f8efb0781713c0c","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN0.acme/anchors","initiator":"f0a0e14476c34242df6a769de4725f3399ba84496c8dd9195dffced05c2ced7c"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN0.acme","sequenceNumber":3,"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"9de8f58234a823c9b9618d852accb42559c0b16ae225f9bdbe9993731a4b6d5c","signature":"a1f4eaef69044c5f43eedcc34150f5fb9d778d241307d6eaf9a49900d5589c286d7b9ce16bed949317d181b895c48681a6f201d1df7c49f6f7e7c4c0c7d2af09","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN1.acme/anchors","initiator":"77d7dd4e8855a911b63b77d42c8379ecf1d792933ec20490c1665ee111f5b011"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN1.acme","sequenceNumber":3,"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"9de8f58234a823c9b9618d852accb42559c0b16ae225f9bdbe9993731a4b6d5c","signature":"2f4620e6e636c16e2f256d1f459ef20abccd155871fe465d3a7a95939122231f07b6bce1d6042c9cc3621be7b513ff8765de76f853e7a3788c3010476892b30c","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN2.acme/anchors","initiator":"4e85f0d164b8df8d6e21586d2e7c5ad11f09c337c32d6da386b420c5c89254df"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN2.acme","sequenceNumber":3,"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"9de8f58234a823c9b9618d852accb42559c0b16ae225f9bdbe9993731a4b6d5c","signature":"ce4786cbb996367c306d81aaf01f7ddc3e418cb11984cad51f975d0848c02ffe7150c4b04f6ee5d6fe2524416b45c9a1c501c8bde911f1e8c4dcf009ccc65405","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"3e13dd70dec6a74e369e383bb7763e780959297c3d96eca68e286e201fe2a6c9"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":3,"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"9de8f58234a823c9b9618d852accb42559c0b16ae225f9bdbe9993731a4b6d5c","signature":"57984d034e6a45cba9d3eed81a12f245a52acd9ba6e7bd888f99ecb29c1216c39aec03ebe387bcea20c6619367a872aa631e6d7688ee2f88546871b4f2aaf108","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"3e13dd70dec6a74e369e383bb7763e780959297c3d96eca68e286e201fe2a6c9"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":3,"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"78641be5bc680f8247265bf2da3cd31afd111c3f59379d80f095986d5fedd07a","signature":"34afe3e66349497ba99c2029dcd302a6636612f091099315a9f85690c892b8ff80054c89dae8088b3d1144aabbb342da43fcbe9a4489a2954e0622893a9d5000","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN0.acme/anchors","initiator":"f0a0e14476c34242df6a769de4725f3399ba84496c8dd9195dffced05c2ced7c"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN0.acme","sequenceNumber":3,"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"78641be5bc680f8247265bf2da3cd31afd111c3f59379d80f095986d5fedd07a","signature":"e3a6ba3efe8d9f3c6775edd10798ed0bd0de450b9b654a1a4198742941bb5b1cd81d7a27fb30c0bc4273c7714cd84c60d60111c48822d7fb6597c13705eb6e03","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN1.acme/anchors","initiator":"77d7dd4e8855a911b63b77d42c8379ecf1d792933ec20490c1665ee111f5b011"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN1.acme","sequenceNumber":3,"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"78641be5bc680f8247265bf2da3cd31afd111c3f59379d80f095986d5fedd07a","signature":"83bf284063dbdb18e30ebaa0f4df5872705b758096d17e6f363704ad4b1d322c80ad0365d3b6a673ff48c72e62e39df439932b5da236bf8a361b97d0d7eeef06","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN2.acme/anchors","initiator":"4e85f0d164b8df8d6e21586d2e7c5ad11f09c337c32d6da386b420c5c89254df"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN2.acme","sequenceNumber":3,"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"78641be5bc680f8247265bf2da3cd31afd111c3f59379d80f095986d5fedd07a","signature":"ed721287209388efa90d6707d55f1d2afe4c45212e83ed20b0f0fd922e9fadd4dcb89d915efa84b7dc3d7fbdc82662273ebe49ef9784dd12150cbfa414868208","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN2.acme/anchors","initiator":"4e85f0d164b8df8d6e21586d2e7c5ad11f09c337c32d6da386b420c5c89254df"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN2.acme","sequenceNumber":3,"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"05f3fcc7e95c17eaab6692d673176249f114a16bc28d2996eb5c08f6c6880320","signature":"663c7c922355f4215b4b06ffca8846bbe365f4a0964f88c32d398ca4be544996be4b69f590f548c3577a77c6f3f7c75d6b43c31d73adc71bcfcb8a1be566020d","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},"transactionHash":"1b0a9de4dc8df5e2cd0993bd420703d50188244c02600e9f98c6ca05fa9234d1"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://dn.acme/anchors","initiator":"3e13dd70dec6a74e369e383bb7763e780959297c3d96eca68e286e201fe2a6c9"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://dn.acme","sequenceNumber":3,"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"05f3fcc7e95c17eaab6692d673176249f114a16bc28d2996eb5c08f6c6880320","signature":"0aea9a309b0eadea2c030d87066b9cf1c6e77fdd9de3fda0280fd9df28a020d150021081c2b5ed8a32b3864e8879395520206ed2fb1f1be8ee4904cc4719ba08","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},"transactionHash":"3ee6a20cd2ef2527f49147508718c0bb2bc00c439174e42ecfb8faaaeafa2743"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN0.acme/anchors","initiator":"f0a0e14476c34242df6a769de4725f3399ba84496c8dd9195dffced05c2ced7c"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN0.acme","sequenceNumber":3,"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"05f3fcc7e95c17eaab6692d673176249f114a16bc28d2996eb5c08f6c6880320","signature":"eef109b2a5262167a3b02eec20e47ff12b429f75ef21ea7100418ac4e1a4d2fb34985a60d850e812af53c0dee3b54662f5b33b22714f5bce39629d1cdcf4a706","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},"transactionHash":"4fdd0d29965c96b06527c163d3d4155e515fdb925464341d21acdd6068867d96"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bvn-BVN1.acme/anchors","initiator":"77d7dd4e8855a911b63b77d42c8379ecf1d792933ec20490c1665ee111f5b011"},"body":{"type":"directoryAnchor","source":"acc://dn.acme","minorBlockIndex":6,"rootChainIndex":26,"rootChainAnchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","stateTreeAnchor":"ffa4f4e2313f76133d51239dd01fc9af14f831460d3b9d2fb1458051b43876cf","receipts":[{"anchor":{"source":"acc://bvn-BVN0.acme","minorBlockIndex":3,"rootChainIndex":18,"rootChainAnchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","stateTreeAnchor":"ffd789d74902c3ed2224c66517578af6c115d1be90f72241b11a97432c6682fb"},"rootChainReceipt":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]}}],"makeMajorBlockTime":"0001-01-01T00:00:00Z"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://dn.acme","destinationNetwork":"acc://bvn-BVN1.acme","sequenceNumber":3,"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"05f3fcc7e95c17eaab6692d673176249f114a16bc28d2996eb5c08f6c6880320","signature":"3eb1aa8325d8cf97d328cd3b69a7f1318e0663585ddd88a7d3bb1427f52a08003e0c6356cf3a329dab6feaeda2fac6431b627771bc40075736e522f03c812c02","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},"transactionHash":"3251a5d2445fe1b9f98c337df190ea8ca6ba8c1f4e98efb51a3ae41fc3ac7fb9"},
	{"type":"transaction","transaction":{"header":{"principal":"acc://bob/tokens","initiator":"2c9f892efd31a8cf7e3e7dd1c32a30258fc99c64cd87afddec8b4a64f297ef7e"},"body":{"type":"syntheticDepositTokens","cause":"acc://0a46b8493ba8f208d37d02bd0c11fe1ac5787c7924b4725f26c4c689e5a6b904@alice/tokens","source":"acc://alice/tokens","initiator":"acc://alice/book/1","feeRefund":100,"token":"acc://ACME","amount":"123"}}},
	{"type":"signature","signature":{"type":"partition","sourceNetwork":"acc://bvn-BVN0.acme","destinationNetwork":"acc://bvn-BVN1.acme","sequenceNumber":1,"transactionHash":"0d96dced01266ecbafdbcf9f8de834ec493936f41fcf6a95121283066b02553a"},"transactionHash":"0d96dced01266ecbafdbcf9f8de834ec493936f41fcf6a95121283066b02553a"},
	{"type":"signature","signature":{"type":"receipt","sourceNetwork":"acc://bvn-BVN0.acme","proof":{"start":"0d96dced01266ecbafdbcf9f8de834ec493936f41fcf6a95121283066b02553a","startIndex":1,"end":"0d96dced01266ecbafdbcf9f8de834ec493936f41fcf6a95121283066b02553a","endIndex":1,"anchor":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","entries":[{"hash":"e43be90e349210456662d8b8bdc9cc9e5e46ccb07f2129e7b57a8195e5e916d5"},{"hash":"eb23de116a2e81eb8496eb05f9205b11a99589201b27fd59673426a46ec5da2b"},{"hash":"37da48d97249200314781ec8e9462ceebc6707232c22e8c4a5822868e2626467"}]},"transactionHash":"0d96dced01266ecbafdbcf9f8de834ec493936f41fcf6a95121283066b02553a"},"transactionHash":"0d96dced01266ecbafdbcf9f8de834ec493936f41fcf6a95121283066b02553a"},
	{"type":"signature","signature":{"type":"receipt","sourceNetwork":"acc://dn.acme","proof":{"start":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","startIndex":1,"end":"a707ba10939db7e1c1c0ce2f638424fcd62ebcc5cf8ca9c5cb71570a9b9f6d35","endIndex":1,"anchor":"bfe648ed21a7e41d6a3f2b58fa7ec7b57f9394a3ae957a6c054ddda9c5165809","entries":[{"hash":"facdad5487838f30e74c395a9589f3403c9c2e19047cfc5b210f507864f09136"},{"hash":"fd8e810e22004038a9c7b8e9dda77fa1c1de25ef3c7005f240aa1566e8c0b87e"},{"right":true,"hash":"aa3d369d90c1078f56fa968e4bf1a17565dc8214ae49941b209a56892e9e7a96"},{"hash":"0847f0e5b2f170a61e1691bc5f531e9fbe182673896e40301777c73c337a2d52"},{"hash":"758fd3b853aca724a2af7cdfbd83fb75a7a2d6282397c6a673c18a54c85b2724"}]},"transactionHash":"0d96dced01266ecbafdbcf9f8de834ec493936f41fcf6a95121283066b02553a"},"transactionHash":"0d96dced01266ecbafdbcf9f8de834ec493936f41fcf6a95121283066b02553a"},
	{"type":"signature","signature":{"type":"ed25519","publicKey":"f0ba4305a861d7a3c1869f2632886e07134a03115f182f860b1113afd0121438","signature":"757f78536060a73f430937d7849bcf6204d9113da0245178644ff1fb3c299b776bbb3b663b5f86ba6759885f3c985778197506ce94ad89d19936bda1b5167f0b","signer":"acc://dn.acme/network","timestamp":1,"transactionHash":"0d96dced01266ecbafdbcf9f8de834ec493936f41fcf6a95121283066b02553a"},"transactionHash":"0d96dced01266ecbafdbcf9f8de834ec493936f41fcf6a95121283066b02553a"}
]`
